// Routinen zur numerischen Optimierung mit dem Extrenen DONLP2-Modul


/* Einschritt-Optimierung 
NOptimize(f,[vars],[ivals])
NOptimize(f,[vars],[ivals],[eqs])
NOptimize(f,[vars],[ivals],[eqs oder ineqs])
*/
NOptimize:=proc()
local i,f,vars,ivals,eqs,Es,Is;
begin
  if args(0)<3 or args(0)>4 then error("Falsche Argumentzahl") end_if; 
  if args(0)=3 then eqs:=[]; else eqs:=args(4); end_if;
  f:=args(1);
  vars:=args(2);
  ivals:=float(args(3));
  if not(domtype(vars)=DOM_LIST) then error("Zweits Argument muss Variablenliste sein!"); end_if;
  if not(domtype(ivals)=DOM_LIST) then error("Drittes  Argument muss Werteliste sein!"); end_if;
  map(ivals, x -> if not(domtype(x)=DOM_FLOAT) then error("FLOAT-Liste fr Startpunkt ntig!") end_if);
  Es:=[]; Is:=[];
  for i from 1 to nops(eqs) do
     if type(eqs[i])="_equal" then Es:=[op(Es),eqs[i]]; else Is:=[op(Is),eqs[i]]; end_if;
  end_for;
  Is:=map(Is, ieq -> if type(ieq)="_less" or type(ieq)="_leequal" then  op(ieq,2)-op(ieq,1) else ieq end_if);
  SetOptim(f,vars,Es,Is,[]);
  NOptim(ivals,[])
end_proc:


DLPSmooth:=TRUE: // Glttung des Absolutbetrages 

SetOptim:=proc(f,vars:DOM_LIST,eqls:DOM_LIST,ineqls:DOM_LIST,params:DOM_LIST)
local i,nv,allvars,grad,el,iel,elg,ielg,n,z,h,g,varstr;
save u;
begin
	delete u;
	allvars:=[op(vars),op(params)];
	for i from 1 to nops(allvars) do
	  if type(allvars[i])="_index" then 
        nv:=makeunindexed(allvars[i]);
		f:=subs(f,allvars[i]=nv); 
		eqls:=subs(eqls,allvars[i]=nv);
		ineqls:=subs(ineqls,allvars[i]=nv);
		allvars[i]:=nv;
 	  end_if;
    end_for;
	if DLPSmooth then
		f:=DLpreprocess(f);
		eqls:=DLpreprocess(eqls);
		ineqls:=DLpreprocess(ineqls);
	end_if;
	vars:=allvars[1..nops(vars)]; 
	params:=allvars[nops(vars)+1..nops(allvars)];
	grad:= ff -> map(vars, v->diff(ff,v)); 
	z:=[f,eqls,ineqls]; 
	map(allvars,v->(z:=subs(z,v=u))); 
	// print("z=",z);
	if not(indets(z)={u} or indets(z)={}) then error("Nicht alle Variablen und Parameter benannt"); end_if;
	varstr:=expr2text(allvars);
	n:=nops(vars);
	varstr:=stringlib::remove(varstr[2..length(varstr)-1]," ");
    el:=map(map(DLpreprocessF(map(eqls,eq2term)),expr2text),LOG2log); 
    elg:=map2(map2(DLpreprocessF(map(map(eqls,eq2term),grad)),expr2text),LOG2log); 
    iel:=map(map(DLpreprocessF(map(ineqls,eq2term)),expr2text),LOG2log); 
    ielg:=map2(map2(DLpreprocessF(map(map(ineqls,eq2term),grad)),expr2text),LOG2log); 
	GlobalDebugInfo:=[LOG2log(expr2text(DLpreprocessF(f))), n, varstr, 
		map(vars, v -> LOG2log(expr2text(DLpreprocessF(diff(f,v))))),
	    el,iel,elg,ielg,nops(params)];
	SetOptim0(LOG2log(expr2text(DLpreprocessF(f))), n, varstr, 
		map(vars, v -> LOG2log(expr2text(DLpreprocessF(diff(f,v))))),
     	el,iel,elg,ielg,nops(params))
end_proc:

eq2term:=proc(e)
begin
  if type(e)="_equal" then op(e,1)-op(e,2) else e end_if;
end_proc:

map2:= (l,f)-> map(l,x->map(x,f)):

makeunindexed:=proc(v)
save F_X;
begin
delete F_X;
if type(v)="_index" then op(v,1)."_".op(v,2).F_X else v end_if;
end_proc:


//SetOptim(x^2+y^2,[x,y],[x^2=y^3],[]);
// NOptim([5,4]);

// Glttung des Betrags !!!!!!!!!!!!!!!!!!!!!!!

DONLPalpha:=1000.0:
DONLPalpha:=10.0:
myabs:= x-> (ln(2)+ln(1+cosh(DONLPalpha*x)))/DONLPalpha:

DLpreprocess:= proc(a)
local u,v;
begin
	if atomicTypeQ(a) then a 
	elif type(a)="abs" then float(myabs(DLpreprocess(op(a,1))))
	elif type(a)="sign" then u:= DLpreprocess(op(a,1)); return(float(u/myabs(u)));
	else map(a,DLpreprocess);
	end_if
end_proc:

DLpreprocessF:= proc(a)
local u,v;
begin
	if atomicTypeQ(a) then a 
	elif type(a)="arcsin" then return(asin(DLpreprocessF(op(a,1)))); 
	elif type(a)="arctan" then return(atan(DLpreprocessF(op(a,1)))); 
	elif type(a)="arccos" then return(acos(DLpreprocessF(op(a,1)))); 
	elif type(a)="ln" then return(LOG(DLpreprocessF(op(a,1)))); 
	elif type(a)="log" then u:= DLpreprocessF(op(a,1));v:= DLpreprocessF(op(a,2));
			return(LOG(v)/LOG(u)); 
	else map(a,DLpreprocessF);
	end_if
end_proc:

LOG2log:= s -> stringlib::subs(s,"LOG"="log"):

atomicTypeQ:=proc(x) begin
  bool(domtype(x)=DOM_INT  or domtype(x)=DOM_COMPLEX or domtype(x)=DOM_BOOL or 
  domtype(x)=DOM_FAIL or domtype(x)=DOM_FLOAT or domtype(x)=DOM_IDENT or 
  domtype(x)=DOM_NIL or domtype(x)=DOM_NULL or domtype(x)=DOM_RAT or 
  domtype(x)=DOM_STRING or domtype(x)=DOM_VAR)  
end_proc:



